/****************************************************************\
**  BB_Sudoku  Bit Based Sudoku Solver                          **
\****************************************************************/

/****************************************************************\
**  (c) Copyright Brian Turner, 2008-2009.  This file may be    **
**      freely used, modified, and copied for personal,         **
**      educational, or non-commercial purposes provided this   **
**      notice remains attached.                                **
\****************************************************************/

#include <stdlib.h>              // For exit()
#include <stdio.h>               // For IO (fgets, stdin, printf)
#include <string.h>
#include "bb_sudoku_tables.h"    // Include the various tables used when solving

// G - Grid data structure
//       This contains everything needed for backtracking.
//       The larger this gets, the slower backtracking will be
//       since the entire structure is copied.
struct Grid_Type
   {
   int      CellsLeft;  // Cells left to solve
   unsigned Grid[81];   // Possibilities left for each cell
   unsigned Grp[27];    // Values found in each group
   };

#define PushSingle(p, b)         \
   {                             \
   SinglePos[SingleCnt] = p;     \
   SingleVal[SingleCnt] = b;     \
   SingleCnt++;                  \
   Changed = 1;                  \
   }

#define MarkChanged(x)           \
   {                             \
   ChangedGroup[C2Grp[x][0]] = ChangedGroup[C2Grp[x][1]] = ChangedGroup[C2Grp[x][2]] = Changed = 1;   \
   }
   
class BritSolver
   {
   private:
      Grid_Type G[128];
      // Solution Grid, which will contain the answer after the puzzle is solved
      unsigned int SolGrid[81];
      // Naked Singles FIFO stack - new singles found are stored here before
      int   SingleCnt;
      char  SinglePos[1024];
      unsigned int SingleVal[1024];
      // Changed Flags
      int  Changed;           // Flag to indicate Grid has changed
      char ChangedGroup[27];  // Marks which groups have changed
      char ChangedLC[27];
      int  SingleGroup[27];
      // Guess structure
      char GuessPos[128];
      unsigned GuessVal[128];
      // Key global variables used throughout the solving
      long  PuzzleCnt;        // Total puzzles tested
      long  SolvedCnt;        // Total puzzles solved
      int   No_Sol;           // Flag to indicate there is no solution possible
      int   PIdx;             // Position Index, used for guessing and backtracking
      int   LastPos;          // used by MakeGuess
   
   public:
      BritSolver() {}

   /************\
   **  Solver  *****************************************************\
   **                                                              **
   **    Solver runs the sudoku solver.  Input puzzle is in the    **
   **    buffer array, and somewhat controlled by a number of      **
   **    globals (see the globals at the top of the main program   **
   **    for globals and meanings).                                **
   **                                                              **
   \****************************************************************/
   int GetNumberOfCompletions( char const* puzzle, unsigned completions[][ 81 ] )
      {
      int PuzSolCnt = 0;

      // init member variables
      SingleCnt = 0;
      PuzzleCnt = 0;
      SolvedCnt = 0;
      LastPos = 0;
      G[ 0 ].CellsLeft = 81;
      PIdx = 0;
      Changed = 1;
      No_Sol = 0;
      // Loop through the buffer and set the singles to be set
      for ( int i = 0;  i < 81;  i++ )
         {
         G[0].Grid[ i ] = 0x1FF;
         SolGrid[ i ] = 0;
         }
      // Clear the Groups Found values
      for ( int i = 0;  i < 27;  i++ )
         G[ 0 ].Grp[ i ] = ChangedGroup[ i ] = ChangedLC[ i ] = SingleGroup[ i ] = 0;
      // add clues
      for ( int c = 0;  c < 81;  c++ )
         if ( puzzle[ c ] >= '1'  &&  puzzle[ c ] <= '9' )
            PushSingle( static_cast< char >( c ), V2B[ puzzle[ c ] - '0' ] );
      // Loop through the puzzle solving routines until finished
      while ( Changed )
         { // If No Solution possible, jump straight to the backtrack routine
         if ( !No_Sol )
            { // Check if any Singles to be propogated
            if ( SingleCnt )
               { 
               if ( SingleCnt > 2 )                // If multiple singles
                  ProcessInitSingles();            // process them all at once
               if ( SingleCnt )                    // otherwise
                  ProcessSingles();                // process them one at a time
               if ( ! G[PIdx].CellsLeft )
                  {
                  if ( ! No_Sol )
                     {           
                     // save solution found
                     memcpy( completions[ PuzSolCnt++ ], SolGrid, 81 * 4 );
                     if ( PuzSolCnt > 1 )
                        return 2;
                     }
                  No_Sol = Changed = 1;
                  continue;
                  }
               }

            // If nothing has changed, apply the next solver
            if ( Changed )
               {
               FindHiddenSingles();
               if ( SingleCnt )
                  continue;
               }
            }

         // If nothing new found, just make a guess
         MakeGuess();
         if ( No_Sol )
            break;
         if ( PIdx > 62 ) { printf( "Max Depth exceeded, recompile for more depth.\n\n" ); exit( 0 ); }
         }

      return PuzSolCnt;
      }

   private:
   /************************\
   **  ProcessInitSingles  *****************************************\
   **                                                              **
   **    ProcessInitSingles takes a naked single and marks each    **
   **    cell in the 3 associated groups as not allowing that      **
   **    number.  It also marks the groups as changed so we know   **
   **    to check for hidden singles in that group.                **
   **                                                              **
   **    This routines marks all the groups first, then marks the  **
   **    cells for each changed groups.                            **
   **                                                              **
   \****************************************************************/
   void ProcessInitSingles()
      {
      while ( SingleCnt > 2 )
         {
         for ( int i = 0;  i < SingleCnt;  i++ )
            {
            int t = SinglePos[ i ];                      // Get local copy of position
            unsigned b = SingleVal[ i ];                 // Get local copy of the value
            if ( G[PIdx].Grid[t] == 0 )                  // Check if we already processed this position
               continue;
            if ( (G[PIdx].Grid[t] & b) == 0 )            // Check for error conditions
               { No_Sol = 1; SingleCnt = Changed = 0; return; }  
            SolGrid[SinglePos[i]] = SingleVal[i];        // Store the single in the solution grid
            G[PIdx].CellsLeft--;                         // mark one less empty space
            G[PIdx].Grid[t] = 0;                         // mark this position processed
            for ( int g = 0;  g < 3;  g++ )              // loop through all 3 groups 
               {
               if ( G[PIdx].Grp[C2Grp[t][g]] & b )
                  { No_Sol = 1; SingleCnt = Changed = 0; return; }  
               G[PIdx].Grp[C2Grp[t][g]] |= b;            // mark the value as found in the group
               SingleGroup[C2Grp[t][g]]  = 1;
               }
            }
         SingleCnt = 0;
         for ( int i = 0;  i < 27;  i++ )
            if ( SingleGroup[i] )
               {
               SingleGroup[i] = 0;
               for ( int j = 0;  j < 9;  j++ )
                  {
                  int t2 = Group[i][j];                  // get temp copy of position
                  unsigned b = G[PIdx].Grp[i];
                  if ( G[PIdx].Grid[t2] & b )            // check if removing a possibility
                     {
                     G[PIdx].Grid[t2] &= ~b;             // remove possibility
                     if ( G[PIdx].Grid[t2] == 0 )        // check for error (no possibility)
                        { No_Sol = 1; SingleCnt = 0; Changed = 0; return; }  
                     if ( B2V[G[PIdx].Grid[t2]] )        // Check if a naked single is found
                        PushSingle( static_cast< char >( t2 ), G[PIdx].Grid[t2] );         
                     MarkChanged(t2);                    // Mark groups as changed
                     }
                  }
               }
         }
      }

   /********************\
   **  ProcessSingles  *********************************************\
   **                                                              **
   **    ProcessSingles takes a naked single and marks each cell   **
   **    in the 3 associated groups as not allowing that number.   **
   **    It also marks the groups as changed so we know to check   **
   **    for hidden singles in that group.                         **
   **                                                              **
   **    This routines marks cells changed as each single is       **
   **    processed.                                                **
   **                                                              **
   \****************************************************************/
   void ProcessSingles()
      {
      for ( int i = 0;  i < SingleCnt;  i++ )      
         {
         int t = SinglePos[i];                     // Get local copy of position
         unsigned b = SingleVal[i];                // Get local copy of the value
         if ( G[PIdx].Grid[t] == 0 )
            continue;                              // Check if we already processed this position
         if ( (G[PIdx].Grid[t] & b) == 0 )         // Check for error conditions
            { No_Sol = 1; SingleCnt = Changed = 0; return; }
         SolGrid[SinglePos[i]] = SingleVal[i];     // Store the single in the solution grid
         G[PIdx].CellsLeft--;                      // mark one less empty space
         G[PIdx].Grid[t] = 0;                      // mark this position processed
         for ( int g = 0;  g < 3;  g++ )           // loop through all 3 groups
            G[PIdx].Grp[C2Grp[t][g]] |= b;         // mark the value as found in the group
         for ( int g = 0;  g < 20;  g++ )          // loop through each cell in the groups
            {
            int t2 = In_Groups[t][g];              // get temp copy of position
            if ( G[PIdx].Grid[t2] & b )            // check if removing a possibility
               {
               G[PIdx].Grid[t2] ^= b;              // remove possibility
               if ( G[PIdx].Grid[t2] == 0 )        // check for error (no possibility)
                  { No_Sol = 1; SingleCnt = 0; Changed = 0; return; }
               if ( B2V[G[PIdx].Grid[t2]] )        // Check if a naked single is found
                  PushSingle( static_cast< char >( t2 ), G[PIdx].Grid[t2]);
               MarkChanged(t2);                    // Mark groups as changed
               }
            }
         }
      SingleCnt = 0;                          // Clear the single count
      }

   /***********************\
   **  FindHiddenSingles  ******************************************\
   **                                                              **
   **    FindHiddenSingles checks each grouping that has changed   **
   **    to see if they contain any hidden singles.  If one is     **
   **    found, the routine adds it to the queue and exits.        **
   **                                                              **
   \****************************************************************/
   void FindHiddenSingles()
      {
      for ( int i = 0;  i < 27;  i++ )
         if ( ChangedGroup[i] )
            {
            ChangedLC[i] = 1;
            unsigned t1 = 0;
            unsigned t2 = 0;
            for ( int j = 0;  j < 9;  j++ )
               {
               unsigned b = G[PIdx].Grid[Group[i][j]];
               t2 |= (t1 & b);
               t1 |= b;
               }
            if ( (t1 | G[PIdx].Grp[i]) != 0x01FF )
               { No_Sol = 1; SingleCnt = 0; Changed = 0; return; }
            unsigned t3 = t2 ^ t1;
            if ( t3 )
               {
               for ( int j = 0;  j < 9;  j++ )
               if ( t3 & G[PIdx].Grid[Group[i][j]] )
                  {
                  unsigned t = t3 & G[PIdx].Grid[Group[i][j]];
                  if ( B2V[t] == 0 )
                     { No_Sol = 1; SingleCnt = 0; Changed = 0; return; }
                  PushSingle( static_cast< char >( Group[i][j] ), t );
                  if ( t3 == t )
                     return;
                  t3 ^= t;
                  }
               }
            ChangedGroup[i] = 0;
            }
      Changed = 0;
      }

   /***************\
   **  MakeGuess  **************************************************\
   **                                                              **
   **    MakeGuess handles the guessing and backtracking used      **
   **    when solving puzzles.                                     **
   **                                                              **
   \****************************************************************/
   void MakeGuess()
      {
      Changed = 1;
      SingleCnt = 0;
      // Check to see where the next guess should be
      while ( No_Sol )
         {
         if ( PIdx == 0 )
            return;
         G[PIdx-1].Grid[GuessPos[PIdx]] ^= GuessVal[PIdx];
         if ( G[PIdx-1].Grid[GuessPos[PIdx]] )
            {
            if ( B2V[G[PIdx-1].Grid[GuessPos[PIdx]]] )
               PushSingle( GuessPos[PIdx], G[PIdx-1].Grid[GuessPos[PIdx]] );
            No_Sol = 0;
            MarkChanged( GuessPos[PIdx] );
            PIdx--;
            return;
            }
         }
      // Populate the grid for the next guess
      G[PIdx+1] = G[PIdx];
      PIdx++;
      // Find next spot to check
      char mt = 99;
      int i = LastPos;
      do
         {
         if ( NSet[G[PIdx].Grid[i]] < mt  &&  NSet[G[PIdx].Grid[i]] > 1 )
            {
            GuessPos[PIdx] = i;
            mt = NSet[G[PIdx].Grid[i]]; 
            if ( mt == 2 )
               break;  // if 2, then its the best it can get
            }
         if ( ++i == 81 )
            i = 0;
         }
      while ( i != LastPos );
      LastPos = i;
      // Mark the next guess in the position
      No_Sol = 1;
      unsigned v = G[PIdx].Grid[GuessPos[PIdx]];
      if ( v )
         {
         v &= -v;    // clear all but the lowest bit set 
         GuessVal[PIdx] = v;
         PushSingle( GuessPos[PIdx], v );
         Changed = 1;
         No_Sol = 0;
         }
      }
   };
