#include <algorithm>
#include <iostream>
#include <fstream>
#include <vector>
#include "binvec.h"
#include "checker.h"
#include "unique.h"
#include "unavsets.h"
#include "uset.h"
#include "util.h"
#include "checker16.h"

using std::cout;
using std::cerr;
using std::endl;
using std::sort;
using std::vector;
using std::ifstream;
using std::min;

#define USETS data
#define FULLVEC data + 24576
#define STATE11 data + 76416
#define STATE12 data + 77056
#define STATE13 data + 77440

// modifying these constants influences the performance
int const START2 = 44;
int const START3 = 30;
int const START4 = 27;
int const START5 = 21;

class UnavMaker
   {  
   public:
      UnavMaker( char const* grid ) : _hitvec( NULL ), _pairhitvec( NULL ), _triplehitvec( NULL ), _quadhitvec( NULL ), _quinthitvec( NULL )
         {
         // find minimal unavoidables
         int A[ 9 ][ 9 ], A_t[ 9 ][ 9 ];
         for ( int i = 0;  i < 81;  i++ )
            {
            A[ i / 9 ][ i % 9 ]   = grid[ i ] - '0';
            A_t[ i % 9 ][ i / 9 ] = A[ i / 9 ][ i % 9 ];
            }
         vector< UnavoidableSet > setvec;
         FindUnavoidableSets( A, setvec );
         FindUnavoidableSets( A_t, setvec, true );
         // make sure that the unavoidables are ordered by size, that there are no duplicates, and that each set is in fact unavoidable
         sort( setvec.begin(), setvec.end() );
         _sets = new UnavoidableSet[ setvec.size() ];
         _numsets = 0;
         for ( unsigned i = 0;  i < setvec.size();  i++ )
            if ( (_numsets == 0  ||  ! (setvec[ i ] == _sets[ _numsets - 1 ]))  &&  setvec[ i ].IsUnavoidable( grid )  &&  setvec[ i ].GetSize() <= 12 )
               _sets[ _numsets++ ] = setvec[ i ];
         // precompute table that records for each pair of unavoidables whether or not the two sets are disjoint
         _disjtab = new bool*[ _numsets ];
         for ( int i = 0;  i < _numsets;  i++ )
            {
            _disjtab[ i ] = new bool[ _numsets ];
            for ( int j = 0;  j < _numsets;  j++ )
               _disjtab[ i ][ j ] = _sets[ i ].IsDisjoint( _sets[ j ] );
            }
         }
      
      ~UnavMaker()
         {
         delete[] _sets;
         delete[] _hitvec;
         delete[] _pairhitvec;
         delete[] _triplehitvec;
         delete[] _quadhitvec;
         delete[] _quinthitvec;
         for ( int i = 0;  i < _numsets;  i++ )
            delete[] _disjtab[ i ];
         delete[] _disjtab;
         }
      
      BinVec384 const* Generate1Hitvec( UnavoidableSet* sets, int* pNumsets ) 
         {
         _hitvec = new BinVec384[ 81 ];
         int const maxsets = min( _numsets, _hitvec[ 0 ].GetLength() );
         for ( int i = 0;  i < 81;  i++ )
            for ( int j = 0;  j < maxsets;  j++ )
               if ( ! _sets[ j ].HasClueFast( i ) )
                  _hitvec[ i ].SetBit( j );
         memcpy( sets, _sets, maxsets * sizeof( UnavoidableSet ) );
         *pNumsets = maxsets;
         return _hitvec;
         }

      BinVec8192 const* Generate2Hitvec()
         {
         _pairhitvec = new BinVec8192[ 81 ];
         int cnt = 0;
         for ( int i = START2;  i < _numsets;  i++ )
            for ( int j = START2 - 1;  j < i;  j++ )
               if ( AreDisj( i, j ) )
                  {
                  // the union of _sets[ i ] and _sets[ j ] is unavoidable of degree 2
                  for ( int c = 0;  c < 81;  c++ )
                     if ( ! _sets[ i ].HasClueFast( c )  &&  ! _sets[ j ].HasClueFast( c ) )
                        _pairhitvec[ c ].SetBit( cnt );
                  if ( ++cnt == _pairhitvec[ 0 ].GetLength() )
                     return _pairhitvec;
                  }
         return _pairhitvec;
         }

      BinVec16384 const* Generate3Hitvec()
         {
         _triplehitvec = new BinVec16384[ 81 ];
         int cnt = 0;
         for ( int i = START3;  i < _numsets;  i++ )
            for ( int j = START3 - 1;  j < i;  j++ )
               {
               if ( ! AreDisj( i, j ) )
                  continue;
               for ( int k = START3 - 2;  k < j;  k++ )
                  if ( AreDisj( i, k )  &&  AreDisj( j, k ) )
                     {
                     // the union of _sets[ i ], _sets[ j ], _sets[ k ] is unavoidable of degree 3
                     UnavoidableSet u( _sets[ i ], _sets[ j ], _sets[ k ] );
                     bool subset = false;
                     for ( int s = 0;  s < START3 - 2;  s++ )
                        if ( u.IsSubset( _sets[ s ] ) )
                           {
                           subset = true;
                           break;
                           }
                     if ( ! subset )
                        {
                        for ( int c = 0;  c < 81;  c++ )
                           if ( ! u.HasClueFast( c ) )
                              _triplehitvec[ c ].SetBit( cnt );
                        if ( ++cnt == _triplehitvec[ 0 ].GetLength() )
                           return _triplehitvec;
                        }
                     }
               }
         return _triplehitvec;
         }

      BinVec32768 const* Generate4Hitvec()
         {
         _quadhitvec = new BinVec32768[ 81 ];
         UnavoidableSet* const sets4 = new UnavoidableSet[ _quadhitvec[ 0 ].GetLength() ];
         int cnt = 0;
         vector< int > idx[ 49 ];
         for ( int i = START4;  i < _numsets;  i++ )
            for ( int j = START4 - 1;  j < i;  j++ )
               {
               if ( ! AreDisj( i, j ) )
                  continue;
               for ( int k = START4 - 2;  k < j;  k++ )
                  {
                  if ( ! AreDisj( i, k )  ||  ! AreDisj( j, k ) )
                     continue;
                  for ( int l = START4 - 3;  l < k;  l++ )
                     {
                     if ( ! AreDisj( i, l )  ||  ! AreDisj( j, l )  ||  ! AreDisj( k, l ) )
                        continue;
                     // the union of _sets[ i ], _sets[ j ], _sets[ k ], _sets[ l ] is unavoidable of degree 4
                     UnavoidableSet u( _sets[ i ], _sets[ j ], _sets[ k ], _sets[ l ] );
                     bool subset = false;
                     for ( int s = 0;  s < START4 - 3;  s++ )
                        if ( u.IsSubset( _sets[ s ] ) )
                           {
                           subset = true;
                           break;
                           }
                     if ( ! subset )
                        {
                        idx[ u.GetSize() ].push_back( cnt );
                        sets4[ cnt ] = u;
                        if ( ++cnt == _quadhitvec[ 0 ].GetLength() )
                           goto full;
                        }
                     }
                  }
               }
full:
         cnt = 0;
         for ( int i = 16;  i < arraysize( idx );  i++ )
            for ( unsigned j = 0;  j < idx[ i ].size();  j++ )
               {
               for ( int c = 0;  c < 81;  c++ )
                  if ( ! sets4[ idx[ i ][ j ] ].HasClueFast( c ) )
                     _quadhitvec[ c ].SetBit( cnt );
               cnt++;
               }
         delete[] sets4;
         return _quadhitvec;
         }

      BinVec16384 const* Generate5Hitvec()
         {
         _quinthitvec = new BinVec16384[ 81 ];
         int const MAX5 = 20000;
         UnavoidableSet* const sets5 = new UnavoidableSet[ MAX5 ];
         int cnt = 0, cnt2 = 0;
         for ( int start5 = START5;  start5 >= 16;  start5-- )
            {
            vector< int > idx[ 61 ];
            for ( int i = start5;  i < _numsets;  i++ )
               for ( int j = start5 - 1;  j < i;  j++ )
                  {
                  if ( ! AreDisj( i, j ) )
                     continue;
                  for ( int k = start5 - 2;  k < j;  k++ )
                     {
                     if ( ! AreDisj( i, k )  ||  ! AreDisj( j, k ) )
                        continue;
                     for ( int l = start5 - 3;  l < k;  l++ )
                        {
                        if ( ! AreDisj( i, l )  ||  ! AreDisj( j, l )  ||  ! AreDisj( k, l ) )
                           continue;
                        for ( int m = start5 - 4;  m < l;  m++ )
                           {
                           if ( ! AreDisj( i, m )  ||  ! AreDisj( j, m )  || ! AreDisj( k, m )  ||  ! AreDisj( l, m ) )
                              continue;
                           // the union of _sets[ i ], _sets[ j ], _sets[ k ], _sets[ l ], _sets[ m ] is unavoidable of degree 5
                           UnavoidableSet u( _sets[ i ], _sets[ j ], _sets[ k ], _sets[ l ], _sets[ m ] );
                           bool subset = false;
                           for ( int s = 0;  s < start5 - 4;  s++ )
                              if ( u.IsSubset( _sets[ s ] ) )
                                 {
                                 subset = true;
                                 break;
                                 }
                           if ( ! subset )
                              {
                              idx[ u.GetSize() ].push_back( cnt );
                              sets5[ cnt ] = u;
                              if ( ++cnt == MAX5 )
                                 goto full;
                              }
                           }
                        }
                     }
                  }           
full:
            for ( int i = 20;  i < arraysize( idx );  i++ )
               for ( unsigned j = 0;  j < idx[ i ].size();  j++ )
                  {
                  for ( int c = 0;  c < 81;  c++ )
                     if ( ! sets5[ idx[ i ][ j ] ].HasClueFast( c ) )
                        _quinthitvec[ c ].SetBit( cnt2 );
                  if ( ++cnt2 == _quinthitvec[ 0 ].GetLength() )
                     goto finished;
                  }
            }
finished:
         delete[] sets5;
         return _quinthitvec;
         }

   private:
      bool AreDisj( int i, int j ) const { return _disjtab[ i ][ j ]; }
   
      int               _numsets;
      UnavoidableSet*   _sets;
      bool**            _disjtab;
      BinVec384*        _hitvec;
      BinVec8192*       _pairhitvec;
      BinVec16384*      _triplehitvec;
      BinVec32768*      _quadhitvec;
      BinVec16384*      _quinthitvec;
   };
   
bool AreMiniRowsEqual( char row1[ 3 ], char row2[ 3 ] )
   {
   if ( row1[ 0 ] == row2[ 1 ] )
      return row1[ 1 ] == row2[ 2 ]  &&  row1[ 2 ] == row2[ 0 ]; 
   else if ( row1[ 0 ] == row2[ 2 ] )
      return row1[ 1 ] == row2[ 0 ]  &&  row1[ 2 ] == row2[ 1 ];
   return false;
   }
   
bool DoesGridHave92UnavbleOfTypeI( char const* grid )
   {
   char sz[ 81 ];
   for ( int t = 0;  t < 2;  t++ )
      {
      if ( t == 0 )
         memcpy( sz, grid, 81 );
      else
         {
         // need the transpose
         for ( int i = 0;  i < 9;  i++ )
            for ( int j = 0;  j < 9;  j++ )
               sz[ i * 9 + j ] = grid[ j * 9 + i ];
         }
      int r[ 3 ];
      for ( int s = 0;  s < 3;  s++ )
         for ( r[ 0 ] = 0;  r[ 0 ] < 3;  r[ 0 ]++ )
            {
            char row1[ 3 ];
            memcpy( row1, sz + r[ 0 ] * 9 + s * 3, 3 );
            for ( r[ 1 ] = 3;  r[ 1 ] < 6;  r[ 1 ]++ )
               {
               char row2[ 3 ];
               memcpy( row2, sz + r[ 1 ] * 9 + s * 3, 3 );
               if ( AreMiniRowsEqual( row1, row2 ) )
                  for ( r[ 2 ] = 6;  r[ 2 ] < 9;  r[ 2 ]++ )
                     {
                     char row3[ 3 ];
                     memcpy( row3, sz + r[ 2 ] * 9 + s * 3, 3 );
                     if ( AreMiniRowsEqual( row1, row3 )  &&  AreMiniRowsEqual( row2, row3 ) )
                        return true;
                     }
               }
            }
      }
   return false;
   }

#define LOOP( n, minvec, minhitvec, pairvec, pairhitvec, triplevec, triplehitvec, quadvec, quadhitvec, quintvec, quinthitvec )   \
   NewUnavoidableSet const* set##n;                                                          \
   {                                                                                         \
   int bestidx = minvec[ n ].GetLowestBitSet();                                              \
   int minsize = unavsets[ bestidx ].GetRealSize( deadcluevec[ n ] );                        \
   for ( int i = bestidx + 1;  i < numsets;  i++ )                                           \
      if ( minvec[ n ].IsBitSet( i )  &&  unavsets[ i ].GetRealSize( deadcluevec[ n ] ) < minsize )   \
         {                                                                                   \
         bestidx = i;                                                                        \
         minsize = unavsets[ i ].GetRealSize( deadcluevec[ n ] );                            \
         }                                                                                   \
   set##n = &unavsets[ bestidx ];                                                            \
   }                                                                                         \
   for ( int i##n = 0;  i##n < set##n->GetSize();  i##n++ )                                  \
      {                                                                                      \
      if ( deadcluevec[ n ].IsBitSet( set##n->GetClue( i##n ) ) )                            \
         continue;                                                                           \
      deadcluevec[ n + 1 ].Or( deadcluevec[ n ],  set##n->_deadcluevec[ i##n ] );            \
      minvec[ n + 1 ]     .And( minvec[ n ],      minhitvec[ set##n->GetClue( i##n ) ] );    \
      pairvec[ n + 1 ]    .And( pairvec[ n ],     pairhitvec[ set##n->GetClue( i##n ) ] );   \
      triplevec[ n + 1 ]  .And( triplevec[ n ],   triplehitvec[ set##n->GetClue( i##n ) ] ); \
      quadvec[ n + 1 ]    .And( quadvec[ n ],     quadhitvec[ set##n->GetClue( i##n ) ] );   \
      if ( minvec[ n + 1 ].AreAllBitsZero() )                                                \
         goto foundsmallhittingset;                                                          \
      quintvec[ n + 1 ]   .And( quintvec[ n ],    quinthitvec[ set##n->GetClue( i##n ) ] );

#define CLOSELOOP( n ) }
      
#define LOOP2( n, minvec, minhitvec, pairvec, pairhitvec, triplevec, triplehitvec, quadvec, quadhitvec, quintvec, quinthitvec )     \
   NewUnavoidableSet const* set##n = reinterpret_cast< NewUnavoidableSet const* >( USETS + FindBestIndex( &minvec[ n ], &deadcluevec[ n ], USETS ) ); \
   for ( int i##n = 0;  i##n < set##n->GetSize();  i##n++ )                                  \
      {                                                                                      \
      if ( deadcluevec[ n ].IsBitSet( set##n->GetClue( i##n ) ) )                            \
         continue;                                                                           \
      deadcluevec[ n + 1 ].Or( deadcluevec[ n ],  set##n->_deadcluevec[ i##n ] );            \
      minvec[ n + 1 ]     .And( minvec[ n ],      minhitvec[ set##n->GetClue( i##n ) ] );    \
      pairvec[ n + 1 ]    .And( pairvec[ n ],     pairhitvec[ set##n->GetClue( i##n ) ] );   \
      triplevec[ n + 1 ]  .And( triplevec[ n ],   triplehitvec[ set##n->GetClue( i##n ) ] ); \
      quadvec[ n + 1 ]    .And( quadvec[ n ],     quadhitvec[ set##n->GetClue( i##n ) ] );   \
      if ( minvec[ n + 1 ].AreAllBitsZero() )                                                \
         goto foundsmallhittingset;                                                          \
      quintvec[ n + 1 ]   .And( quintvec[ n ],    quinthitvec[ set##n->GetClue( i##n ) ] );


inline void ConsolidateHitvec( unsigned char* const mask, int const size1, char const* const hitvec1, int const len1,
                              int const size2, char* const hitvec2, int const len2, BinVec128 const& deadcluevec )
   {
   __declspec( align( 16 ) ) char shift[ 4097 ];   // 4097 == 32768 / 8 + 1
   shift[ 0 ] = 0;
   __declspec( align( 16 ) ) int LOOP_MAX[ 24 ];   // 24 == 1536 / 64
   __declspec( align( 16 ) ) int idx = 0;
   for ( int j = 1;  j <= (len1 >> 3);  j++ )
      {
      shift[ j ] = shift[ j - 1 ] + wt[ mask[ j - 1 ] ];
      if ( shift[ j ] >= 64 )
         {
         LOOP_MAX[ idx++ ] = j;
         shift[ j ] -= 64;
         if ( idx == (len2 >> 6) )
            break;
         }
      }
   if ( idx == (len2 >> 6) )
      idx--;
   else
      LOOP_MAX[ idx ] = (len1 >> 3);

   char const* in[ 84 ]; // 84 == ceiling( 81 / 6 ) * 6
   char* out[ 84 ];
   int cnt = 0;
   for ( int i = 0;  i < 81;  i++ )
      if ( ! deadcluevec.IsBitSet( i ) )
         {
         in[ cnt ]  = hitvec1 + i * size1;
         out[ cnt ] = hitvec2 + i * size2;
         cnt++;
         }
   if ( cnt == 0 )
      return;              // only happens when deadcluevec[0..80] == 11..1 (81 1's)
   // now make sure that 'cnt' is evenly divisibly by 6
   while ( (cnt % 6) != 0 )
      {
      in[ cnt ]  = hitvec1 + (cnt % 6) * size1;
      out[ cnt ] = hitvec2 + (cnt % 6) * size2;
      cnt++;
      }
   unsigned char const* p = &confunctab[ 0 ];
   __declspec( align( 4 ) ) unsigned max;
   __asm
      {
      mov   eax, cnt
      mov   rdi, p
      mov   rdx, mask
      pinsrq xmm7, rdi, 1                          // hiqword( xmm7 ) == &confunctab[ 0 ]
      pinsrq xmm7, rdx, 0                          // loqword( xmm7 ) == &mask[ 0 ]
      loop_1:
         mov   r10, [in + rax * 8 - 48]
         mov   r11, [in + rax * 8 - 40]
         mov   r12, [in + rax * 8 - 32]
         mov   r13, [in + rax * 8 - 24]
         mov   r14, [in + rax * 8 - 16]
         mov   r15, [in + rax * 8 - 8]
         movq  xmm8, rax
         xor   edi, edi
         movq  mm0, [out + rax * 8 - 48]
         movq  xmm10, r10
         movq  mm1, [out + rax * 8 - 40]
         movq  xmm11, r11
         movq  mm2, [out + rax * 8 - 32]
         movq  xmm12, r12
         pxor  xmm0, xmm0
         movq  mm3, [out + rax * 8 - 24]
         movq  xmm13, r13
         pxor  xmm1, xmm1
         movq  mm4, [out + rax * 8 - 16]
         movq  xmm14, r14
         pxor  xmm2, xmm2
         movq  mm5, [out + rax * 8 - 8]
         movq  xmm15, r15
         xor   esi, esi                                           // j := 0
         loop_2:
            mov   eax, [LOOP_MAX + rdi * 4]
            movq  mm6, rdi
            cmp   esi, eax                                        // j == LOOP_MAX[ m ]?
            je    done_3
            mov   max, eax
            loop_3:
               movq  rdi, xmm7                                    // rdi == &mask[ 0 ]
               movzx eax, byte ptr [r10 + rsi]                    // rbx := data0
               pextrq rdx, xmm7, 1                                // rdx == &confunctab[ 0 ]
               movzx edi, byte ptr [rdi + rsi]                    // rdi := mask[ j ]
               shl   edi, 8                                       // rdi == mask[ j ] * 256
               jz    next
               movzx ecx, byte ptr [shift + rsi]                  // rcx := shift[ j ]
               add   rdi, rdx                                     // rdi == &confunctab[ mask[ j ] * 256 ]
               movzx ebx, byte ptr [r11 + rsi]                    // rbx := data1
               movq  xmm6, rcx                                    // xmm6 == shift[ j ]
               movzx eax, byte ptr [rdi + rax]                    // rax := confunctab[ data0 + mask[ j ] * 256 ]
               movq  xmm3, rax
               movzx ebx, byte ptr [rdi + rbx]                    // rax := confunctab[ data0 + mask[ j ] * 256 ]
               pinsrq xmm3, rbx, 1
               movzx ecx, byte ptr [r12 + rsi]                    // rcx := data2
               psllq xmm3, xmm6
               movzx edx, byte ptr [r13 + rsi]                    // rdx := data3
               por   xmm0, xmm3
               movzx ecx, byte ptr [rdi + rcx]                    // rcx := confunctab[ data2 + mask[ j ] * 256 ]
               movq  xmm4, rcx
               movzx edx, byte ptr [rdi + rdx]                    // rdx := confunctab[ data3 + mask[ j ] * 256 ]
               pinsrq xmm4, rdx, 1
               movzx r8d, byte ptr [r14 + rsi]                    // r8 := data4
               psllq xmm4, xmm6
               movzx r9d, byte ptr [r15 + rsi]                    // r9 := data5
               por   xmm1, xmm4
               movzx r8d, byte ptr [rdi + r8]                     // r8 := confunctab[ data4 + mask[ j ] * 256 ]
               movq  xmm5, r8
               movzx r9d, byte ptr [rdi + r9]                     // r9 := confunctab[ data5 + mask[ j ] * 256 ]
               pinsrq xmm5, r9, 1
               psllq xmm5, xmm6
               por   xmm2, xmm5
            next:
               add   esi, 1                                       // j++
               cmp   esi, max
               jne   loop_3
         done_3:
            movq  rdi, mm6                                        // restore old value of rdi
            movq  r10, mm0
            movq  r11, mm1
            movq  [r10 + rdi * 8], xmm0
            movq  r10, xmm10
            movq  r12, mm2
            movq  r13, mm3
            pextrq [r11 + rdi * 8], xmm0, 1
            movq  r11, xmm11
            movq  r14, mm4
            movq  r15, mm5
            movq  [r12 + rdi * 8], xmm1
            movq  r12, xmm12
            pextrq [r13 + rdi * 8], xmm1, 1
            movq  r13, xmm13
            pxor  xmm0, xmm0
            movq  [r14 + rdi * 8], xmm2
            movq  r14, xmm14
            pxor  xmm1, xmm1
            pextrq [r15 + rdi * 8], xmm2, 1
            movq  r15, xmm15           
            pxor  xmm2, xmm2

            cmp edi, idx
            je  done_2

            add edi, 1
            cmp byte ptr [shift + rsi], 0       // shift[ j ] == 0 ?
            je  cont
         
            mov   r10, rcx
            mov   ecx, 64
            movq  r11, xmm6                     // r11 == shift[ j - 1 ]            
            sub   ecx, r11d
            movq  r11, xmm11           
            shr   rax, cl
            shr   r10, cl           
            shr   r8, cl
            movq  xmm0, rax
            movq  xmm1, r10            
            shr   rbx, cl
            shr   rdx, cl
            movq  xmm2, r8
            pinsrq xmm0, rbx, 1
            shr   r9, cl
            movq  r10, xmm10           
            pinsrq xmm1, rdx, 1
            pinsrq xmm2, r9, 1
            
         cont:
            cmp   edi, idx
            jbe   loop_2
done_2:
         movq  rax, xmm8
         sub   eax, 6
         jnz   loop_1
      }
   }

#define CONSOLIDATE_HITVEC( vec, hitvec, n )                                                             \
   {                                                                                                     \
   vec##2[ n ].SetAllBitsTo1();                                                                          \
   ConsolidateHitvec( reinterpret_cast< unsigned char* >( vec##1[ n ]._bits ),                           \
                      sizeof( vec##1[ 0 ] ), reinterpret_cast< char const* >( hitvec##1[ 0 ]._bits ),    \
                      vec##1[ 0 ].GetLength(),                                                           \
                      sizeof( vec##2[ 0 ] ), reinterpret_cast< char* >( hitvec##2[ 0 ]._bits ),          \
                      vec##2[ 0 ].GetLength(), deadcluevec[ n ] );                                       \
   }

#define CLUE( n ) partpuzz[ set##n->GetClue( i##n ) ] = grid[ set##n->GetClue( i##n ) ];

#define CLUE0TO9                                \
   memset( partpuzz, '0', sizeof( partpuzz ) ); \
   partpuzz[ 81 ] = '\0';                       \
   CLUE( 0 )                                    \
   CLUE( 1 )                                    \
   CLUE( 2 )                                    \
   CLUE( 3 )                                    \
   CLUE( 4 )                                    \
   CLUE( 5 )                                    \
   CLUE( 6 )                                    \
   CLUE( 7 )                                    \
   CLUE( 8 )                                    \
   CLUE( 9 )

inline int FindBestIndex( void* minvec, void* deadcluevec, void* usets )
   {
   int idx192;
   __asm
      {
      mov      r14, minvec
      mov      rsi, deadcluevec
      pcmpeqb  xmm1, xmm1

      mov      rax, [r14]
      mov      r15, usets
      mov      r8d, 15

      mov      r9, [r14 + 08h]
      add      r15, 0B0h

      pxor     xmm1, [rsi]
      cmp      rax, 0
      jz       next
align 16
loop1:
      bsf      rcx, rax
      movdqa   xmm0, xmm1
      mov      ebx, 1

      shl      rbx, cl
      shl      ecx, 6

      lea      ecx, [ecx * 2 + ecx]
      not      rbx

      pand     xmm0, [r15 + rcx]
      and      rax, rbx

      movq     rdx, xmm0
      pextrq   rdi, xmm0, 1

      popcnt   rdx, rdx
      popcnt   edi, edi

      add      edi, edx

      cmp      edi, r8d
      cmovb    esi, ecx
      cmovb    r8d, edi

      cmp      rax, 0
      jne      loop1

next:
      mov      rax, r9
      cmp      rax, 0
      jz       done
align 16
loop2:
      bsf      rcx, rax
      movdqa   xmm0, xmm1
      mov      ebx, 1

      shl      rbx, cl
      shl      ecx, 6

      lea      ecx, [ecx * 2 + ecx + 64 * 192]
      not      rbx

      pand     xmm0, [r15 + rcx]
      and      rax, rbx

      movq     rdx, xmm0
      pextrq   rdi, xmm0, 1

      popcnt   rdx, rdx
      popcnt   edi, edi

      add      edi, edx

      cmp      edi, r8d
      cmovb    esi, ecx
      cmovb    r8d, edi

      cmp      rax, 0
      jne      loop2

done:
      mov      idx192, esi 
      }
   return idx192;
   }

// #define CACHELINE 64
// #define HALFCACHELINE 32

#define CACHELINE 16
#define HALFCACHELINE 16


int CheckGrid( char const* grid, float* pGridtime )
   {
   BinVec128   deadcluevec[ 12 ];
   BinVec384   minvec1[ 8 ];
   BinVec8192  pairvec1[ 7 ];
   BinVec16384 triplevec1[ 7 ];
   BinVec32768 quadvec1[ 6 ];
   BinVec16384 quintvec1[ 5 ];
   BinVec128   onemask[ 81 ];
   BinVec128   allmask[ 128 ];
   __declspec( align( CACHELINE ) ) char data[ 77632 ];
   int counter = 0;
   float starttime = GetCPUTime();
   // acquire the unavoidable sets for this grid
   UnavMaker unavmaker( grid );
   // set up hitting vectors for the higher-degree unavoidables
   BinVec8192 const*  pairhitvec1   = unavmaker.Generate2Hitvec();
   BinVec16384 const* triplehitvec1 = unavmaker.Generate3Hitvec();
   BinVec32768 const* quadhitvec1   = unavmaker.Generate4Hitvec();
   BinVec16384 const* quinthitvec1  = unavmaker.Generate5Hitvec();
   // set up hitting vectors for the minimal unavoidables
   UnavoidableSet* unavsetsold = new UnavoidableSet[ 384 ];
   int numsets;
   BinVec384 const* hitvec1 = unavmaker.Generate1Hitvec( unavsetsold, &numsets );
   // put the hitting vectors in the state req'd such that we can perform the actual enumeration of hitting sets
   NewUnavoidableSet* unavsets = new NewUnavoidableSet[ 384 ];
   for ( int i = 0;  i < numsets;  i++ )
      {
      unavsetsold[ i ].Finalize();
      unavsets[ i ] = unavsetsold[ i ];
      }
   delete[] unavsetsold;
   // 81 binary vectors of length 128, where the i-th vector has a 1 in only the i-th slot
   for ( int i = 0;  i < 81;  i++ )
      {
      onemask[ i ].SetBit( i );
      memcpy( FULLVEC + i * 640 + 12 * 16, &onemask[ i ], 16 );
      memcpy( FULLVEC + i * 640 + 25 * 16, &onemask[ i ], 16 );
      }
   // 128 binary vectors of length 128, where the i-th vector has 1's in slot 0 .. (i-1)
   for ( int i = 1;  i < 128;  i++ )
      {
      allmask[ i ] = allmask[ i - 1 ];
      allmask[ i ].SetBit( i - 1 );
      }
   // find all hitting sets of size 16 and run them through the solver to see if there are any with a unique completion
   // initially, no unavoidable set is hit
   minvec1[ 0 ].SetAllBitsTo1();
   pairvec1[ 0 ].SetAllBitsTo1();
   triplevec1[ 0 ].SetAllBitsTo1();
   quadvec1[ 0 ].SetAllBitsTo1();
   quintvec1[ 0 ].SetAllBitsTo1();
   short* const puzzarrorg = new short[ 4 * (1048576 + 65536) + 32 ];
   // make sure puzzarr is CACHELINE aligned
   short* const puzzarr = ( short* )((reinterpret_cast< uint64 >( puzzarrorg ) + 63) & ~0x3F);
   // enumerate hitting sets
   bool found16 = false;
   LOOP( 0, minvec1, hitvec1, pairvec1, pairhitvec1, triplevec1, triplehitvec1, quadvec1, quadhitvec1, quintvec1, quinthitvec1 )
   LOOP( 1, minvec1, hitvec1, pairvec1, pairhitvec1, triplevec1, triplehitvec1, quadvec1, quadhitvec1, quintvec1, quinthitvec1 )
   LOOP( 2, minvec1, hitvec1, pairvec1, pairhitvec1, triplevec1, triplehitvec1, quadvec1, quadhitvec1, quintvec1, quinthitvec1 )
   LOOP( 3, minvec1, hitvec1, pairvec1, pairhitvec1, triplevec1, triplehitvec1, quadvec1, quadhitvec1, quintvec1, quinthitvec1 )
      // discard all but the smallest 1536 unavoidables of degree 5 that are not yet hit
      __declspec( align( CACHELINE ) ) BinVec1536 quintvec2[ 12 ];
      __declspec( align( CACHELINE ) ) BinVec1536 quinthitvec2[ 81 ];
      CONSOLIDATE_HITVEC( quintvec, quinthitvec, 4 );
      for ( int i = 0;  i < 81;  i++ )
         if ( ! deadcluevec[ 4 ].IsBitSet( i ) )
            memcpy( FULLVEC + i * 640, &quinthitvec2[ i ], sizeof( quintvec2[ 0 ] ) );
   LOOP( 4, minvec1, hitvec1, pairvec1, pairhitvec1, triplevec1, triplehitvec1, quadvec1, quadhitvec1, quintvec2, quinthitvec2 )
      // discard all but the smallest 1536 unavoidables of degree 4 that are not yet hit
      __declspec( align( CACHELINE ) ) BinVec1536 quadvec2[ 12 ];
      __declspec( align( CACHELINE ) ) BinVec1536 quadhitvec2[ 81 ];
      CONSOLIDATE_HITVEC( quadvec, quadhitvec, 5 );      
      for ( int i = 0;  i < 81;  i++ )
         if ( ! deadcluevec[ 5 ].IsBitSet( i ) )
            memcpy( FULLVEC + i * 640 + 13 * 16, &quadhitvec2[ i ], sizeof( quadvec2[ 0 ] ) );
   LOOP( 5, minvec1, hitvec1, pairvec1, pairhitvec1, triplevec1, triplehitvec1, quadvec2, quadhitvec2, quintvec2, quinthitvec2 )
      // discard all but the smallest 1280 unavoidables of degree 3 that are not yet hit
      __declspec( align( HALFCACHELINE ) ) BinVec1280 triplevec2[ 12 ];
      __declspec( align( HALFCACHELINE ) ) BinVec1280 triplehitvec2[ 81 ];
      CONSOLIDATE_HITVEC( triplevec, triplehitvec, 6 );      
      // similar for the unavoidables of degree 2 (retain only the smallest 384)
      __declspec( align( CACHELINE ) ) BinVec384 pairvec2[ 12 ];
      __declspec( align( CACHELINE ) ) BinVec384 pairhitvec2[ 81 ];
      CONSOLIDATE_HITVEC( pairvec, pairhitvec, 6 );      
      for ( int i = 0;  i < 81;  i++ )
         if ( ! deadcluevec[ 6 ].IsBitSet( i ) )
            {
            memcpy( FULLVEC + i * 640 + 26 * 16, &triplehitvec2[ i ], sizeof( triplevec2[ 0 ] ) );
            memcpy( FULLVEC + i * 640 + 36 * 16, &pairhitvec2[ i ], 48 );
            }
         LOOP( 6, minvec1, hitvec1, pairvec2, pairhitvec2, triplevec2, triplehitvec2, quadvec2, quadhitvec2, quintvec2, quinthitvec2 )
         // discard all but the smallest 128 minimal unavoidables that are not yet hit
         BinVec128 minvec2[ 12 ];
         BinVec128 hitvec2[ 81 ];
         CONSOLIDATE_HITVEC( minvec, hitvec, 7 );
         for ( int i = 0;  i < 81;  i++ )
            if ( ! deadcluevec[ 7 ].IsBitSet( i ) )
               memcpy( FULLVEC + i * 640 + 39 * 16, &hitvec2[ i ], 16 );
         {
         int k = 0;
         for ( int j = 0;  j < numsets  &&  k < minvec2[ 0 ].GetLength();  j++ )
            if ( minvec1[ 7 ].IsBitSet( j ) )
               memcpy( USETS + 192 * k++, &unavsets[ j ], 192 );
         if ( k != minvec2[ 0 ].GetLength() )
            memcpy( &minvec2[ 7 ], &allmask[ k ], 16 );
         }

         LOOP2( 7, minvec2, hitvec2, pairvec2, pairhitvec2, triplevec2, triplehitvec2, quadvec2, quadhitvec2, quintvec2, quinthitvec2 )
         LOOP2( 8, minvec2, hitvec2, pairvec2, pairhitvec2, triplevec2, triplehitvec2, quadvec2, quadhitvec2, quintvec2, quinthitvec2 )
         LOOP2( 9, minvec2, hitvec2, pairvec2, pairhitvec2, triplevec2, triplehitvec2, quadvec2, quadhitvec2, quintvec2, quinthitvec2 )

         __declspec( align( 4 ) ) int puzztmp = 0;
         __asm
            {
            mov r12, puzzarr        // r12 = &puzzarr[ 0 ]
            xor r14d, r14d          // r14 = puzzcnt
            // copy minvec2[ 10 ] to the end of pairvec2[ 10 ]
            movdqa   xmm0, [minvec2 + 10 * 16]
            pcmpeqb  xmm1, xmm1
            xor      r11d, r11d
            mov      ecx, 64 * 192
            mov      r9d, 15
            lea      r15, [USETS + 0B0h]
            movq     rax, xmm0
            pextrq   rbx, xmm0, 1
            movdqa   [pairvec2 + 10 * 64 + 30h], xmm0
            pxor     xmm1, [deadcluevec + 10 * 16]                // xmm1 := alivecluevec[ 10 ]
            cmp      rax, 0
            cmovz    r11d, ecx
            cmovz    rax, rbx
            add      r15, r11
align 16
pickusetloop:
            bsf      rcx, rax
            movdqa   xmm0, xmm1
            mov      ebx, 1

            shl      rbx, cl
            shl      ecx, 6

            lea      ecx, [ecx * 2 + ecx]
            not      rbx

            pand     xmm0, [r15 + rcx]             // clues alive in this unavoidable set
            and      rax, rbx

            movq     rdx, xmm0
            pextrq   rdi, xmm0, 1

            popcnt   rdx, rdx
            add      ecx, r11d
            popcnt   edi, edi

            add      edi, edx                      // edi == real size of this unavoidable set

            cmp      edi, r9d
            cmovb    esi, ecx
            cmovb    r9d, edi

            cmp      rax, 0
            jne pickusetloop

            // rax := minvec2[ 10 ].GetLowestBitSet();
            lea      r10, [USETS + rsi + 0Eh]         // r10 = &set10 + 14
            movzx    r8d, word ptr [r10 + 0Eh]        // r8 := set10.GetSize() << 4
align 16
loop10:
               movzx    eax, word ptr [r10]                 // rbx == clue10 * 128
               add      r10, 10h

               movdqa   xmm6, [deadcluevec + 10 * 16]       // xmm6 := deadcluevec[ 10 ]
               lea      ebx, [eax * 4 + eax]                // rbx == clue10 * 640
               shr      eax, 3                              // rax == clue10 * 16
               sub      r8d, 10h

               movdqa   xmm5, xmm6                          // check if clue10 is still alive

               pand     xmm6, [onemask + rax]
               shr      eax, 4                              // rax == clue10

               ptest    xmm6, xmm6
               jnz      next10

                  // deadcluevec[ 11 ] := deadcluevec[ 10 ] | set10._deadcluevec[ i10 ]
                  por      xmm5, [r10 - 1Eh]                
                  movq     mm7, rax                            // save clue10 to bits 0-7 of mm7
                  lea      rax, [quintvec2 + 10 * 192]         // rax := &quintvec2[ 10 ]
                  // quintvec2[ 11 ] := quintvec2[ 10 ] & quinthitvec2[ clue10 ]
                  movdqa   xmm0, [rax]
                  lea      rbx, [FULLVEC + rbx]                // rbx == &FULLVEC[ clue10 ]
                  movdqa   xmm1, [rax + 10h]
                  lea      rdx, [STATE11]
                  pand     xmm0, [rbx]
                  movdqa   [rdx], xmm5
                  add      rdx, 10h
                  pand     xmm1, [rbx + 10h]
                  movdqa   xmm2, [rax + 20h]
                  movdqa   xmm3, [rax + 30h]
                  movdqa   [rdx], xmm0
                  movdqa   [rdx + 10h], xmm1
                  pand     xmm2, [rbx + 20h]
                  pand     xmm3, [rbx + 30h]
                  movdqa   [rdx + 20h], xmm2
                  movdqa   [rdx + 30h], xmm3

                  movdqa   xmm0, [rax + 40h]
                  movdqa   xmm1, [rax + 50h]
                  pand     xmm0, [rbx + 40h]
                  pand     xmm1, [rbx + 50h]
                  movdqa   xmm2, [rax + 60h]
                  movdqa   xmm3, [rax + 70h]
                  add      rax, 70h
                  movdqa   [rdx + 40h], xmm0
                  movdqa   [rdx + 50h], xmm1
                  pand     xmm2, [rbx + 60h]
                  pand     xmm3, [rbx + 70h]
                  add      rbx, 70h
                  movdqa   [rdx + 60h], xmm2
                  movdqa   [rdx + 70h], xmm3
                  add      rdx, 70h

                  movdqa   xmm0, [rax + 10h]
                  movdqa   xmm1, [rax + 20h]
                  pand     xmm0, [rbx + 10h]
                  pand     xmm1, [rbx + 20h]
                  movdqa   xmm2, [rax + 30h]
                  movdqa   xmm3, [rax + 40h]
                  movdqa   [rdx + 10h], xmm0
                  movdqa   [rdx + 20h], xmm1
                  pand     xmm2, [rbx + 30h]
                  lea      rax, [quadvec2 + 10 * 192]
                  pand     xmm3, [rbx + 40h]
                  add      rbx, 60h
                  movdqa   [rdx + 30h], xmm2
                  movdqa   [rdx + 40h], xmm3
                  add      rdx, 50h

                  // quadvec2[ 11 ] = quadvec2[ 10 ] & quadhitvec2[ clue10 ]
                  movdqa   xmm0, [rax]
                  movdqa   xmm1, [rax + 10h]
                  pand     xmm0, [rbx]
                  pand     xmm1, [rbx + 10h]
                  movdqa   xmm2, [rax + 20h]
                  movdqa   xmm3, [rax + 30h]
                  movdqa   [rdx], xmm0
                  movdqa   [rdx + 10h], xmm1
                  pand     xmm2, [rbx + 20h]
                  pand     xmm3, [rbx + 30h]
                  movdqa   [rdx + 20h], xmm2
                  movdqa   [rdx + 30h], xmm3

                  movdqa   xmm0, [rax + 40h]
                  movdqa   xmm1, [rax + 50h]
                  pand     xmm0, [rbx + 40h]
                  pand     xmm1, [rbx + 50h]
                  movdqa   xmm2, [rax + 60h]
                  movdqa   xmm3, [rax + 70h]
                  add      rax, 70h
                  movdqa   [rdx + 40h], xmm0
                  movdqa   [rdx + 50h], xmm1
                  pand     xmm2, [rbx + 60h]
                  pand     xmm3, [rbx + 70h]
                  add      rbx, 70h
                  movdqa   [rdx + 60h], xmm2
                  movdqa   [rdx + 70h], xmm3
                  add      rdx, 70h

                  movdqa   xmm0, [rax + 10h]
                  movdqa   xmm1, [rax + 20h]
                  pand     xmm0, [rbx + 10h]
                  pand     xmm1, [rbx + 20h]
                  movdqa   xmm2, [rax + 30h]
                  movdqa   xmm3, [rax + 40h]
                  movdqa   [rdx + 10h], xmm0
                  movdqa   [rdx + 20h], xmm1 
                  pand     xmm2, [rbx + 30h]
                  lea      rax, [triplevec2 + 10 * 160]
                  pand     xmm3, [rbx + 40h]
                  add      rbx, 60h
                  movdqa   [rdx + 30h], xmm2
                  movdqa   [rdx + 40h], xmm3
                  add      rdx, 50h

                  // triplevec2[ 11 ] := triplevec2[ 10 ] & triplehitvec2[ clue10 ]
                  movdqa   xmm0, [rax]
                  movdqa   xmm1, [rax + 10h]
                  pand     xmm0, [rbx]
                  pand     xmm1, [rbx + 10h]
                  movdqa   xmm2, [rax + 20h]
                  movdqa   xmm3, [rax + 30h]
                  movdqa   [rdx], xmm0
                  movdqa   [rdx + 10h], xmm1 
                  pand     xmm2, [rbx + 20h]
                  pand     xmm3, [rbx + 30h]
                  movdqa   [rdx + 20h], xmm2
                  movdqa   [rdx + 30h], xmm3 

                  movdqa   xmm0, [rax + 40h]
                  movdqa   xmm1, [rax + 50h]
                  pand     xmm0, [rbx + 40h]
                  pand     xmm1, [rbx + 50h]
                  movdqa   xmm2, [rax + 60h]
                  movdqa   xmm3, [rax + 70h]
                  add      rax, 70h
                  movdqa   [rdx + 40h], xmm0
                  movdqa   [rdx + 50h], xmm1 
                  pand     xmm2, [rbx + 60h]
                  pand     xmm3, [rbx + 70h]
                  add      rbx, 70h
                  movdqa   [rdx + 60h], xmm2
                  movdqa   [rdx + 70h], xmm3
                  add      rdx, 70h

                  movdqa   xmm0, [rax + 10h]
                  movdqa   xmm1, [rax + 20h]
                  pand     xmm0, [rbx + 10h]
                  lea      rax, [pairvec2 + 10 * 64]
                  pand     xmm1, [rbx + 20h]
                  add      rbx, 30h
                  movdqa   [rdx + 10h], xmm0
                  movdqa   [rdx + 20h], xmm1
                  add      rdx, 30h
                  // pairvec2[ 11 ] := pairvec2[ 10 ] & pairhitvec2[ clue10 ]
                  movdqa   xmm0, [rax]
                  movdqa   xmm1, [rax + 10h]
                  pand     xmm0, [rbx]
                  pand     xmm1, [rbx + 10h]
                  movdqa   xmm2, [rax + 20h]
                  movdqa   xmm3, [rax + 30h]  // this is really minvec2[ 10 ]
                  movdqa   [rdx], xmm0
                  movdqa   [rdx + 10h], xmm1
                  pand     xmm2, [rbx + 20h]
                  pand     xmm3, [rbx + 30h]
                  pcmpeqb  xmm1, xmm1
                  movdqa   [rdx + 20h], xmm2
                  movdqa   [rdx + 30h], xmm3
                  movq     rax, xmm3
                  pextrq   rbx, xmm3, 1

                  xor      r11d, r11d
                  mov      ecx, 64 * 192
                  mov      r9d, 15
                  lea      r15, [USETS + 0B0h]
                  pxor     xmm1, xmm5              // xmm1 == alivevec[ 11 ] (xmm5 == deadcluevec[ 11 ])
                  cmp      rax, 0
                  cmovz    r11d, ecx
                  cmovz    rax, rbx
                  add      r15, r11

                  cmp      rax, 0
                  jz       foundsmallhittingset

                  bsf      rcx, rax
                  movdqa   xmm0, xmm1
                  mov      ebx, 1
                  shl      rbx, cl
                  shl      ecx, 6
                  lea      ecx, [ecx * 2 + ecx]
                  not      rbx
                  pand     xmm0, [r15 + rcx]
                  and      rax, rbx
                  movq     rdx, xmm0
                  pextrq   rdi, xmm0, 1
                  popcnt   rdx, rdx
                  add      ecx, r11d
                  popcnt   edi, edi
                  add      edi, edx
                  cmp      edi, r9d
                  cmovb    esi, ecx
                  cmovb    r9d, edi

                  popcnt   rbx, rax
                  cmp      ebx, 4
                  jb       foundbit11

                  bsf      rcx, rax
                  movdqa   xmm0, xmm1
                  mov      ebx, 1
                  shl      rbx, cl
                  shl      ecx, 6
                  lea      ecx, [ecx * 2 + ecx]
                  not      rbx
                  pand     xmm0, [r15 + rcx]
                  and      rax, rbx
                  movq     rdx, xmm0
                  pextrq   rdi, xmm0, 1
                  popcnt   rdx, rdx
                  add      ecx, r11d
                  popcnt   edi, edi
                  add      edi, edx
                  cmp      edi, r9d
                  cmovb    esi, ecx
                  cmovb    r9d, edi

                  bsf      rcx, rax
                  movdqa   xmm0, xmm1
                  mov      ebx, 1
                  shl      rbx, cl
                  shl      ecx, 6
                  lea      ecx, [ecx * 2 + ecx]
                  not      rbx
                  pand     xmm0, [r15 + rcx]
                  and      rax, rbx
                  movq     rdx, xmm0
                  pextrq   rdi, xmm0, 1
                  popcnt   rdx, rdx
                  add      ecx, r11d
                  popcnt   edi, edi
                  add      edi, edx
                  cmp      edi, r9d
                  cmovb    esi, ecx
                  cmovb    r9d, edi

                  bsf      rcx, rax
                  movdqa   xmm0, xmm1
                  mov      ebx, 1
                  shl      rbx, cl
                  shl      ecx, 6
                  lea      ecx, [ecx * 2 + ecx]
                  not      rbx
                  pand     xmm0, [r15 + rcx]
                  and      rax, rbx
                  movq     rdx, xmm0
                  pextrq   rdi, xmm0, 1
                  popcnt   rdx, rdx
                  add      ecx, r11d
                  popcnt   edi, edi
                  add      edi, edx
                  cmp      edi, r9d
                  cmovb    esi, ecx
                  cmovb    r9d, edi

                  bsf      rcx, rax
                  movdqa   xmm0, xmm1
                  mov      ebx, 1
                  shl      rbx, cl
                  shl      ecx, 6
                  lea      ecx, [ecx * 2 + ecx]
                  not      rbx
                  pand     xmm0, [r15 + rcx]
                  and      rax, rbx
                  movq     rdx, xmm0
                  pextrq   rdi, xmm0, 1
                  popcnt   rdx, rdx
                  add      ecx, r11d
                  popcnt   edi, edi
                  add      edi, edx
                  cmp      edi, r9d
                  cmovb    esi, ecx
foundbit11:
                  lea      r11, [USETS + rsi + 0Eh]                  // r11 := &set11 + 0Eh
                  movzx    ebx, word ptr [r11 - 02h]                 // ebx := set11.GetSize()
                  add      r8d, ebx
align 16
loop11:
                     movzx    ebx, word ptr [r11]                       // rbx := clue11 << 7
                     lea      rdx, [STATE11 + 10h]
                     add      r11, 10h

                     movdqa   xmm0, [rdx - 10h]                         // xmm0 := deadcluevec[ 11 ]
                     lea      rax, [FULLVEC + rbx * 4]                  // rax + rbx == &FULLVEC[ clue11 ]

                     movdqa   xmm1, [rdx]
                     add      rax, rbx

                     movdqa   xmm2, [rdx + 10h]

                     pand     xmm1, [rax]
                     movdqa   xmm7, xmm0
                     pand     xmm2, [rax + 10h]
                     movdqa   xmm3, [rdx + 20h]
                     movdqa   xmm4, [rdx + 30h]
                     por      xmm1, xmm2
                     pand     xmm3, [rax + 20h]
                     pand     xmm4, [rax + 30h]
                     pand     xmm0, [rax + 0C0h]                        // xmm0 &= onemask[ clue11 ]
                     add      rax, 40h
                     por      xmm1, xmm3
                     add      rdx, 40h
                     por      xmm1, xmm4
                     sub      r8d, 1
                     por      xmm0, xmm1
                     ptest    xmm0, xmm0
                     jnz      next11

                     movdqa   xmm0, [rdx]
                     movdqa   xmm1, [rdx + 10h]
                     pand     xmm0, [rax]
                     pand     xmm1, [rax + 10h]
                     movdqa   xmm2, [rdx + 20h]
                     movdqa   xmm3, [rdx + 30h]
                     por      xmm0, xmm1
                     pand     xmm2, [rax + 20h]
                     por      xmm0, xmm2
                     pand     xmm3, [rax + 30h]
                     por      xmm0, xmm3
                     ptest    xmm0, xmm0
                     jnz      next11

                     movdqa   xmm0, [rdx + 40h]
                     movdqa   xmm1, [rdx + 50h]
                     pand     xmm0, [rax + 40h]
                     pand     xmm1, [rax + 50h]
                     movdqa   xmm2, [rdx + 60h]
                     movdqa   xmm3, [rdx + 70h]
                     add      rdx, 80h
                     pand     xmm2, [rax + 60h]
                     por      xmm0, xmm1
                     pand     xmm3, [rax + 70h]
                     por      xmm0, xmm2
                     add      rax, 90h
                     por      xmm0, xmm3
                     ptest    xmm0, xmm0
                     jnz      next11

                        movdqa   xmm14, xmm7
                        // quadvec2[ 12 ] := quadvec2[ 11 ] & quadhitvec2[ clue11 ]
                        movdqa   xmm15, [rdx]
                        movq2dq  xmm5, mm7

                        por      xmm14, [r11 - 1Eh]            // xmm14 == deadcluevec[ 12 ]
                        shr      ebx, 7

                        pand     xmm15, [rax]
                        pinsrb   xmm5, ebx, 1

                        movdqa   xmm0, [rdx + 10h]
                        movdqa   xmm1, [rdx + 20h]
                        pand     xmm0, [rax + 10h]
                        movdq2q  mm7, xmm5                     // bits 8-15 of mm7 now contain clue11
                        lea      rbx, [STATE12]
                        pand     xmm1, [rax + 20h]
                        movdq2q  mm0, xmm0
                        movdqa   xmm2, [rdx + 30h]
                        psrldq   xmm0, 8
                        movdqa   xmm3, [rdx + 40h]
                        movdq2q  mm1, xmm0
                        pand     xmm2, [rax + 30h]
                        movdq2q  mm2, xmm1
                        pand     xmm3, [rax + 40h]
                        psrldq   xmm1, 8
                        movdqa   xmm0, [rdx + 50h]
                        movdq2q  mm3, xmm1
                        movdqa   [rbx], xmm3
                        movdqa   xmm1, [rdx + 60h]
                        add      rdx, 70h
                        movdq2q  mm4, xmm2
                        pand     xmm0, [rax + 50h]
                        psrldq   xmm2, 8
                        pand     xmm1, [rax + 60h]
                        add      rax, 70h
                        movdq2q  mm5, xmm2
                        movdqa   [rbx + 10h], xmm0
                        movdqa   [rbx + 20h], xmm1
                        add      rbx, 30h

                        movdqa   xmm2, [rdx]
                        movdqa   xmm3, [rdx + 10h]
                        pand     xmm2, [rax]
                        pand     xmm3, [rax + 10h]
                        movdqa   xmm0, [rdx + 20h]
                        movdqa   [rbx], xmm2
                        movdqa   xmm1, [rdx + 30h]
                        movdqa   [rbx + 10h], xmm3
                        movdqa   xmm2, [rdx + 40h]
                        pand     xmm0, [rax + 20h]
                        pand     xmm1, [rax + 30h]
                        movdqa   [rbx + 20h], xmm0
                        pand     xmm2, [rax + 40h]
                        movdqa   [rbx + 30h], xmm1
                        movdqa   [rbx + 40h], xmm2
                        // triplevec2[ 12 ] := triplevec2[ 11 ] & triplehitvec2[ clue11 ]
                        movdqa   xmm0, [rdx + 50h]
                        movdqa   xmm1, [rdx + 60h]
                        add      rdx, 70h
                        pand     xmm0, [rax + 60h]
                        pand     xmm1, [rax + 70h]
                        movdqa   [rbx + 50h], xmm0
                        add      rax, 80h
                        movdqa   [rbx + 60h], xmm1
                        add      rbx, 70h
                        movdqa   xmm0, [rdx]
                        movdqa   xmm1, [rdx + 10h]
                        pand     xmm0, [rax]
                        pand     xmm1, [rax + 10h]
                        movdqa   xmm2, [rdx + 20h]
                        movdqa   [rbx], xmm0
                        movdqa   xmm3, [rdx + 30h]
                        movdqa   [rbx + 10h], xmm1
                        pand     xmm2, [rax + 20h]
                        pand     xmm3, [rax + 30h]
                        movdqa   xmm0, [rdx + 40h]
                        movdqa   [rbx + 20h], xmm2
                        movdqa   xmm1, [rdx + 50h]
                        movdqa   [rbx + 30h], xmm3
                        pand     xmm0, [rax + 40h]
                        pand     xmm1, [rax + 50h]
                        movdqa   xmm2, [rdx + 60h]
                        movdqa   [rbx + 40h], xmm0
                        movdqa   xmm3, [rdx + 70h]
                        add      rdx, 70h
                        movdqa   [rbx + 50h], xmm1
                        pand     xmm2, [rax + 60h]
                        pand     xmm3, [rax + 70h]
                        add      rax, 70h
                        movdqa   [rbx + 60h], xmm2
                        movdqa   [rbx + 70h], xmm3
                        add      rbx, 70h
                        // pairvec2[ 12 ] := pairvec2[ 11 ] & pairhitvec2[ clue11 ]
                        // minvec2[ 12 ] := minvec1[ 11 ] & minhitvec2[ clue11 ]
                        movdqa   xmm0, [rdx + 10h]
                        movdqa   xmm1, [rdx + 20h]
                        pand     xmm0, [rax + 10h]
                        pand     xmm1, [rax + 20h]
                        movdqa   xmm2, [rdx + 30h]
                        movdqa   [rbx + 10h], xmm0
                        movdqa   xmm3, [rdx + 40h]
                        movdqa   [rbx + 20h], xmm1
                        pand     xmm2, [rax + 30h]
                        pand     xmm3, [rax + 40h]
                        movdqa   [rbx + 30h], xmm2
                        movq     rax, xmm3
                        movdqa   [rbx + 40h], xmm3
                        pextrq   rbx, xmm3, 1
                        bsf      rax, rax
                        jnz      foundbit12
                        bsf      rax, rbx
                        jz       foundsmallhittingset
                        add      eax, 64
align 16
foundbit12:
                        shl      eax, 6

                        lea      rdx, [USETS + rax * 2 + 0Eh]                 // rdx + rax == &set12 + 0Eh

                        movzx    ecx, word ptr [rdx + rax + 1Eh]              // rcx := set12.GetSize() << 8
                        add      rdx, rax
align 16
loop12:
                           movzx    ebx, word ptr [rdx]                       // rdx := clue12 << 7
                           add      rdx, 10h
                           movdqa   xmm0, xmm14
                           movq2dq  xmm2, mm0

                           mov      eax, ebx
                           lea      rbx, [FULLVEC + rbx * 4 + 0C0h]           // rax + rbx == &FULLVEC[ clue12 ] + 12 * 16
                           movq2dq  xmm5, mm1
                           movdqa   xmm1, xmm15

                           pand     xmm0, [rbx + rax]
                           add      rbx, rax
                           pslldq   xmm5, 8
                           movq2dq  xmm3, mm2

                           pand     xmm1, [rbx + 10h]
                           por      xmm2, xmm5
                           movq2dq  xmm6, mm3

                           pand     xmm2, [rbx + 20h]
                           pslldq   xmm6, 8
                           movq2dq  xmm4, mm4

                           por      xmm3, xmm6
                           por      xmm0, xmm1
                           movq2dq  xmm7, mm5

                           pand     xmm3, [rbx + 30h]
                           pslldq   xmm7, 8
                           por      xmm0, xmm2

                           por      xmm4, xmm7
                           sub      ecx, 100h
                           por      xmm0, xmm3

                           pand     xmm4, [rbx + 40h]
                           por      xmm0, xmm4
                           mov      esi, eax                                  // esi == clue12 << 7

                           lea      rax, [STATE12]
                           ptest    xmm0, xmm0
                           jnz      next12

                           movdqa   xmm0, [rax]
                           movdqa   xmm1, [rax + 10h]
                           pand     xmm0, [rbx + 50h]
                           pand     xmm1, [rbx + 60h]
                           add      rbx, 70h
                           movdqa   xmm2, [rax + 20h]
                           movdqa   xmm3, [rax + 30h]
                           por      xmm0, xmm1
                           pand     xmm2, [rbx]
                           por      xmm0, xmm2
                           pand     xmm3, [rbx + 10h]
                           por      xmm0, xmm3
                           ptest    xmm0, xmm0
                           jnz      next12

                           movdqa   xmm0, [rax + 40h]
                           movdqa   xmm1, [rax + 50h]
                           pand     xmm0, [rbx + 20h]
                           pand     xmm1, [rbx + 30h]
                           movdqa   xmm2, [rax + 60h]
                           movdqa   xmm3, [rax + 70h]
                           por      xmm0, xmm1
                           pand     xmm2, [rbx + 40h]
                           por      xmm0, xmm2
                           pand     xmm3, [rbx + 50h]
                           por      xmm0, xmm3
                           ptest    xmm0, xmm0
                           jnz      next12

                              add      rax, 80h
                              // stop if puzzcnt already greater than 1,048,576
                              cmp      r14d, 1048576
                              ja       foundsmallhittingset
                              movdqa   xmm4, xmm14
                              // triplevec2[ 13 ] := triplevec2[ 12 ] & triplehitvec2[ clue12 ]
                              movdqa   xmm5, [rax]

                              add      rbx, 70h                                  // rbx == &FULLVEC[ clue12 ] + 26 * 16
                              por      xmm4, [rdx - 1Eh]                         // xmm4 == deadcluevec[ 13 ]
                              movq2dq  xmm0, mm7

                              pand     xmm5, [rbx]
                              shr      esi, 7

                              movdqa   xmm6, [rax + 10h]
                              pinsrb   xmm0, esi, 2
                              movdqa   xmm7, [rax + 20h]
                              pand     xmm6, [rbx + 10h]
                              pand     xmm7, [rbx + 20h]
                              movdq2q  mm7, xmm0                                 // bits 16-23 of mm7 now contain clue12

                              movdqa   xmm8, [rax + 30h]
                              movdqa   xmm0, [rax + 40h]
                              lea      rdi, [STATE13]
                              pand     xmm8, [rbx + 30h]
                              pand     xmm0, [rbx + 40h]

                              movdqa   xmm1, [rax + 50h]
                              movdqa   xmm2, [rax + 60h]
                              add      rax, 70h
                              pand     xmm1, [rbx + 50h]
                              movdqa   [rdi], xmm0
                              pand     xmm2, [rbx + 60h]
                              add      rbx, 70h
                              movdqa   [rdi + 10h], xmm1
                              movdqa   [rdi + 20h], xmm2

                              movdqa   xmm0, [rax]
                              movdqa   xmm1, [rax + 10h]
                              pand     xmm0, [rbx]
                              pand     xmm1, [rbx + 10h]
                              movdqa   xmm2, [rax + 20h]
                              movdqa   [rdi + 30h], xmm0
                              pand     xmm2, [rbx + 20h]
                              movdqa   [rdi + 40h], xmm1
                              movdqa   [rdi + 50h], xmm2
                              add      rdi, 60h

                              // pairvec, minvec
                              movdqa   xmm3, [rax + 60h]
                              movdqa   xmm0, [rax + 30h]
                              pand     xmm3, [rbx + 60h]
                              pand     xmm0, [rbx + 30h]
                              movdqa   [rdi + 30h], xmm3
                              movdqa   [rdi], xmm0
                              movdqa   xmm1, [rax + 40h]
                              movdqa   xmm2, [rax + 50h]
                              movq     rax, xmm3
                              pand     xmm1, [rbx + 40h]
                              pand     xmm2, [rbx + 50h]
                              pextrq   rbx, xmm3, 1
                              movdqa   [rdi + 10h], xmm1
                              movdqa   [rdi + 20h], xmm2

                              bsf      rax, rax
                              jnz      foundbit13
                              bsf      rax, rbx
                              jz       hittingset13
                              add      eax, 64
align 16
foundbit13:
                              shl      eax, 6

                              lea      rsi, [USETS + rax * 2 + 0Eh]                    // rsi + rax == &set13 + 0Eh

                              movzx    ebx, word ptr [rsi + rax + 0Eh]                 // ecx += set13.GetSize() << 4
                              add      rsi, rax
                              add      ecx, ebx
align 16
loop13:
                                 movzx    ebx, word ptr [rsi]                          // rbx := clue13 << 7
                                 add      rsi, 10h
                                 movdqa   xmm0, xmm4

                                 lea      rax, [FULLVEC + rbx * 4 + 190h]              // rax + rbx == &FULLVEC[ clue13 ] + 25 * 16
                                 movdqa   xmm1, xmm5
                                 movdqa   xmm2, xmm6

                                 pand     xmm0, [rax + rbx]
                                 add      rax, rbx
                                 movdqa   xmm3, xmm7

                                 pand     xmm1, [rax + 10h]
                                 sub      ecx, 10h

                                 pand     xmm2, [rax + 20h]
                                 por      xmm0, xmm1
                                 movdqa   xmm1, xmm8

                                 pand     xmm3, [rax + 30h]
                                 por      xmm0, xmm2

                                 pand     xmm1, [rax + 40h]
                                 por      xmm0, xmm3

                                 lea      rdi, [STATE13]
                                 por      xmm0, xmm1

                                 ptest    xmm0, xmm0
                                 jz       continue13

                                 test     cl, 0F0h
                                 jnz      loop13
                                 jmp      next12
align 16
continue13:
                                 movdqa   xmm0, [rdi]
                                 movdqa   xmm1, [rdi + 10h]
                                 pand     xmm0, [rax + 50h]
                                 pand     xmm1, [rax + 60h]
                                 add      rax, 70h
                                 movdqa   xmm2, [rdi + 20h]
                                 movdqa   xmm3, [rdi + 30h]
                                 por      xmm0, xmm1
                                 pand     xmm2, [rax]
                                 pand     xmm3, [rax + 10h]
                                 por      xmm0, xmm2
                                 movdqa   xmm1, [rdi + 40h]
                                 movdqa   xmm2, [rdi + 50h]
                                 por      xmm0, xmm3
                                 pand     xmm1, [rax + 20h]
                                 pand     xmm2, [rax + 30h]
                                 por      xmm0, xmm1
                                 por      xmm0, xmm2
                                 ptest    xmm0, xmm0
                                 jnz      next13

                                    movq2dq  xmm0, mm7
                                    movdqa   xmm9, xmm4
                                    movdqa   xmm10, [rdi + 60h]
                                    shr      ebx, 7
                                    movdqa   xmm11, [rdi + 70h]
                                    add      rdi, 70h
                                    por      xmm9, [rsi - 1Eh]
                                    pinsrb   xmm0, ebx, 3
                                    pand     xmm10, [rax + 40h]
                                    pand     xmm11, [rax + 50h]
                                    movdqa   xmm12, [rdi + 10h]
                                    movdqa   xmm13, [rdi + 20h]
                                    movdq2q  mm7, xmm0                                 // bit 24-31 of mm7 now contain clue13
                                    pand     xmm12, [rax + 60h]
                                    pand     xmm13, [rax + 70h]
                                    movq     rax, xmm13
                                    pextrq   rbx, xmm13, 1

                                    bsf      rax, rax
                                    jnz      foundbit14
                                    bsf      rax, rbx
                                    jz       hittingset14
                                    add      eax, 64
align 16
foundbit14:
                                    shl      eax, 6

                                    lea      rdi, [USETS + rax * 2 + 0Eh]                    // rdi + rax == &set14 + 0Eh

                                    movzx    ebx, word ptr [rdi + rax - 02h]
                                    add      rdi, rax
                                    add      ecx, ebx                                        // ecx += set14.GetSize()
align 16
loop14:
                                    movzx    ebx, word ptr [rdi]                             // rbx := clue14 << 7
                                    add      rdi, 10h
                                    movdqa   xmm0, xmm9

                                    lea      rax, [FULLVEC + rbx * 4 + 190h]
                                    movdqa   xmm1, xmm10
                                    movdqa   xmm2, xmm11

                                    pand     xmm0, [rax + rbx]
                                    lea      rax, [rax + rbx + 70h]
                                    movdqa   xmm3, xmm12

                                    pand     xmm1, [rax + 40h]
                                    sub      ecx, 1

                                    pand     xmm2, [rax + 50h]
                                    por      xmm0, xmm1

                                    pand     xmm3, [rax + 60h]
                                    por      xmm0, xmm2

                                    por      xmm0, xmm3

                                    ptest    xmm0, xmm0
                                    jz       continue14

                                    test     cl, 0Fh
                                    jnz      loop14
                                    jmp      next13
align 16
continue14:
                                       movdqa   xmm2, xmm9
                                       movdqa   xmm3, xmm13
                                       por      xmm2, [rdi - 1Eh]
                                       pinsrw   mm7, ebx, 2
                                       pand     xmm3, [rax + 70h]
                                       movq     rax, xmm3
                                       pextrq   rbx, xmm3, 1

                                       bsf      rax, rax
                                       jnz      foundbit15
                                       bsf      rax, rbx
                                       jz       hittingset15
                                       add      eax, 64
align 16
foundbit15:
                                       shl      eax, 6

                                       lea      r15, [USETS + rax * 2 + 0Eh]                    // rdx == minvec2[ 14 ].GetLowestBitSet() * 192

                                       movzx    ebx, word ptr [r15 + rax + 2Eh]                 // ecx += set14.GetSize() << 12
                                       add      r15, rax
                                       add      ecx, ebx
loop15:
                                          movzx    ebx, word ptr [r15]                             // ebx := clue15 << 7
                                          add      r15, 10h
                                          movdqa   xmm0, xmm2
                                          lea      r9d, [r14d + 1]

                                          mov      eax, ebx
                                          lea      rbx, [FULLVEC + rbx * 4 + 190h]
                                          movdqa   xmm1, xmm3

                                          pand     xmm0, [rax + rbx]
                                          pinsrw   mm7, eax, 3
                                          add      eax, 0E0h

                                          pand     xmm1, [rax + rbx]
                                          sub      ecx, 01000h

                                          por      xmm0, xmm1
                                          movq     [r12 + r14 * 8], mm7
                                          ptest    xmm0, xmm0
                                          cmovz    r14d, r9d

                                          test     ecx, 0F000h
                                          jnz      loop15
                                    test cl, 0Fh
                                    jnz loop14
                                 test cl, 0F0h
                                 jnz loop13
                              jmp next12
align 16
hittingset15:
                                    xor eax, eax
                                    lea rbx, [onemask]
                                    mov r15d, 81 * 128
addlastclue:
                                       movdqa xmm0, [rbx]                                 // xmm0 := onemask[ clue15 ]
                                       pinsrw mm7, eax, 3
                                       lea   r9d, [r14d + 1]
                                       add   eax, 80h

                                       add   rbx, 10h
                                       pand  xmm0, xmm2                                   // xmm2 == deadcluevec[ 15 ]

                                       movq  [r12 + r14 * 8], mm7
                                       ptest xmm0, xmm0

                                       cmovz r14d, r9d
                                       cmp   eax, r15d
                                       jne   addlastclue
                                    test cl, 0Fh
                                    jnz loop14
                                 jmp next13
align 16
hittingset14:
                                 // found a 14-hitting set, need two more clues
                                 lea r15, [onemask]
                                 xor edi, edi
                                 movdqa xmm1, xmm9                                        // xmm1 := deadcluevec[ 14 ]
                                 movq mm6, r8
                                 mov r8d, 81 * 128
addclue1:
                                 movdqa xmm0, [r15]
                                 pinsrw mm7, edi, 2
                                 add   edi, 80h

                                 pand xmm0, xmm1
                                 ptest xmm0, xmm0                          // check if the clue edi is already dead
                                 jnz clue14dead
                                    mov eax, edi
                                    lea rbx, [r15 + 10h]
addclue2:
                                    movdqa xmm0, [rbx]
                                    add    rbx, 10h
                                    pinsrw mm7, eax, 3
                                    lea    r9d, [r14d + 1]

                                    pand   xmm0, xmm1
                                    movq  [r12 + r14 * 8], mm7
                                    add   eax, 80h
                                    ptest  xmm0, xmm0                      // check if eax is dead
                                    cmovz r14d, r9d
                                    cmp   eax, r8d
                                    jne   addclue2
clue14dead:
                                 add   r15, 10h
                                 cmp   edi, 80 * 128
                                 jne   addclue1
                                 movq r8, mm6
                                 test cl, 0F0h
                                 jnz loop13
                              jmp next12
align 16
hittingset13:
                           // found a 13-hitting set, need three more clues
                           movq     mm6, rcx
                           mov      ecx, 81 * 128
                           movq2dq  xmm1, mm7
                           lea      r15, [onemask]
                           pinsrw   mm6, r8d, 1
                           xor      r8d, r8d
align 16
addclue1of3:
                              movdqa xmm0, [r15]
                              pinsrb xmm1, r8d, 3
                              add    r8d, 1
                              pand   xmm0, xmm4
                              ptest  xmm0, xmm0                      // check if the clue r8 is already dead
                              jnz clue13dead
                              mov   edi, r8d
                              lea   rsi, [r15 + 10h]
                              shl   edi, 7
addclue2of3:
                                 movdqa xmm0, [rsi]
                                 pinsrw xmm1, edi, 2
                                 add   edi, 80h
                                 pand  xmm0, xmm4
                                 ptest xmm0, xmm0                   // check if rcx/16 is dead
                                 jnz   clue14dead2
                                 lea   rax, [rsi + 10h]
                                 mov   ebx, edi
addclue3of3:
                                    movdqa   xmm0, [rax]
                                    pinsrw   xmm1, ebx, 3
                                    add      ebx, 80h
                                    pand     xmm0, xmm4
                                    lea      r9d, [r14d + 1]
                                    movdqu   [r12 + r14 * 8], xmm1
                                    ptest    xmm0, xmm0                // check if rdx/16 is dead
                                    cmovz    r14d, r9d
                                    add      rax, 10h
                                    cmp      ebx, ecx
                                    jne      addclue3of3
clue14dead2:
                                 add   rsi, 10h
                                 cmp   edi, 80 * 128
                                 jne addclue2of3
align 16
clue13dead:
                              add   r15, 10h
                              cmp   r8d, 79
                              jne addclue1of3
                           pextrw ecx, mm6, 0
                           pextrw r8d, mm6, 1
                           jmp next12
align 16
next13:
                                 test cl, 0F0h
                                 jnz loop13
align 16
next12:
                           test ecx, 0F00h
                           jnz loop12
align 16
next11:
                     test r8b, 0Fh
                     jnz loop11
align 16
next10:
               test r8b, 0F0h
               jnz loop10

            mov puzztmp, r14d
            }
         if ( puzztmp > 0 )
            {
            __declspec( align( HALFCACHELINE ) ) char partpuzz[ 82 ];
            CLUE0TO9;
            counter += puzztmp;
            for ( int i = 0;  i < puzztmp;  i++ )
               {
               char puzzle[ 82 ];
               memcpy( puzzle, partpuzz, sizeof( puzzle ) );
               int clue11 = puzzarr[ i * 4 ] & 0xFF;
               int clue12 = puzzarr[ i * 4 ] >> 8;
               int clue13 = puzzarr[ i * 4 + 1 ] & 0xFF;
               int clue14 = puzzarr[ i * 4 + 1 ] >> 8;
               int clue15 = puzzarr[ i * 4 + 2 ] >> 7;
               int clue16 = puzzarr[ i * 4 + 3 ] >> 7;
               puzzle[ clue11 ] = grid[ clue11 ];
               puzzle[ clue12 ] = grid[ clue12 ];
               puzzle[ clue13 ] = grid[ clue13 ];
               puzzle[ clue14 ] = grid[ clue14 ];
               puzzle[ clue15 ] = grid[ clue15 ];
               puzzle[ clue16 ] = grid[ clue16 ];
#ifndef NDEBUG
               int cnt = 0;
               for ( int i = 0;  i < 81;  i++ )
                  if ( puzzle[ i ] != '0' )
                     cnt++;
               assert( cnt == 16 );
#endif
               int const comp = GetNumberOfCompletionsOfPuzzle( puzzle );
               if ( comp == 1 )
                  {
                  found16 = true;
                  std::cerr << "*********** Found a 16!!! ***********" << std::endl
                            << puzzle << std::endl << grid << std::endl
                            << "*************************************" << std::endl;
                  std::ofstream foundFile( "FOUND_A_16", std::ios::app );
                  foundFile << puzzle << std::endl << grid << std::endl;
                  }                                                                                   
               else if ( comp == -1 )
                  goto solver_problem;
               }
            }
         CLOSELOOP( 9 )
         CLOSELOOP( 8 )
         CLOSELOOP( 7 )
      CLOSELOOP( 6 )
      CLOSELOOP( 5 )
   CLOSELOOP( 4 )
   CLOSELOOP( 3 )
   CLOSELOOP( 2 )
   CLOSELOOP( 1 )
   CLOSELOOP( 0 )
   __asm emms
   if ( pGridtime != NULL )
      *pGridtime = GetCPUTime() - starttime;
   delete[] unavsets;
   delete[] puzzarrorg;
   return (found16 ? -1 : counter);

   solver_problem:
   // seems there is a problem with the solver, log this grid
   std::cerr << "Solver problem encountered while searching this grid: " << grid << std::endl;
   __asm emms;
   if ( pGridtime != NULL )
      *pGridtime = 0.0;
   delete[] unavsets;
   delete[] puzzarrorg;
   return -2;  

   foundsmallhittingset:
   // hitting sets of size less than 13 are not currently handled
   // record any grid having hitting sets of size 12 or below
   std::cerr << "Found a grid with a small hitting set: " << grid << std::endl;
   std::ofstream specialFile( "grids_with_small_hitting_set.txt", std::ios::app );
   specialFile << grid << std::endl;
   __asm emms;
   if ( pGridtime != NULL )
      *pGridtime = 0.0;
   delete[] unavsets;
   delete[] puzzarrorg;
   return -3;
   }

bool DoesGridHave92UnavbleOfTypeII( char const* grid )
   {
   char  sz[ 81 ];
   int   s, t;
   int   r[ 3 ];
   for ( t = 0;  t < 2;  t++ )
      {
      if ( t == 0 )
         memcpy( sz, grid, 81 );
      else
         {
         // need the transpose
         for ( int i = 0;  i < 9;  i++ )
            for ( int j = 0;  j < 9;  j++ )
               sz[ i * 9 + j ] = grid[ j * 9 + i ];
         }
      char rows[ 3 ][ 3 ];
      for ( s = 0;  s < 3;  s++ )
         for ( r[ 0 ] = 0;  r[ 0 ] < 3;  r[ 0 ]++ )
            {
            memcpy( rows[ 0 ], sz + r[ 0 ] * 9 + s * 3, 3 );
            for ( r[ 1 ] = 3;  r[ 1 ] < 6;  r[ 1 ]++ )
               {
               memcpy( rows[ 1 ], sz + r[ 1 ] * 9 + s * 3, 3 );
                  for ( r[ 2 ] = 6;  r[ 2 ] < 9;  r[ 2 ]++ )
                     {
                     memcpy( rows[ 2 ], sz + r[ 2 ] * 9 + s * 3, 3 );
                     if ( rows[ 0 ][ 0 ] == rows[ 1 ][ 1 ]  &&  rows[ 0 ][ 1 ] == rows[ 1 ][ 0 ] )
                        {
                        // check type i)
                        // abc
                        // bad
                        // cda
                        if ( rows[ 1 ][ 1 ] == rows[ 2 ][ 2 ]  &&  rows[ 0 ][ 2 ] == rows[ 2 ][ 0 ]  &&  rows[ 1 ][ 2 ] == rows[ 2 ][ 1 ] )
                           return true;
                        // check type ii)
                        // abc
                        // bad
                        // dcb
                        if ( rows[ 1 ][ 0 ] == rows[ 2 ][ 2 ]  &&  rows[ 0 ][ 2 ] == rows[ 2 ][ 1 ]  &&  rows[ 1 ][ 2 ] == rows[ 2 ][ 0 ] )
                           return true;
                        }
                     else if ( rows[ 0 ][ 0 ] == rows[ 2 ][ 1 ]  &&  rows[ 0 ][ 1 ] == rows[ 2 ][ 0 ] )
                        {
                        // check type iii)
                        // abc
                        // cda                        
                        // bad
                        if ( rows[ 0 ][ 2 ] == rows[ 1 ][ 0 ]  &&  rows[ 1 ][ 1 ] == rows[ 2 ][ 2 ]  &&  rows[ 0 ][ 0 ] == rows[ 1 ][ 2 ] )
                           return true;
                        // check type iv)
                        // abc
                        // dcb                        
                        // bad
                        else if ( rows[ 0 ][ 2 ] == rows[ 1 ][ 1 ]  &&  rows[ 0 ][ 1 ] == rows[ 1 ][ 2 ]  &&  rows[ 1 ][ 0 ] == rows[ 2 ][ 2 ] )
                           return true;
                        }
                     else if ( rows[ 1 ][ 0 ] == rows[ 2 ][ 1 ]  &&  rows[ 1 ][ 1 ] == rows[ 2 ][ 0 ] )
                        {
                        // check type v)
                        // cda                        
                        // abc
                        // bad
                        if ( rows[ 0 ][ 0 ] == rows[ 1 ][ 2 ]  &&  rows[ 0 ][ 1 ] == rows[ 2 ][ 2 ]  &&  rows[ 0 ][ 2 ] == rows[ 1 ][ 0 ] )
                           return true;
                        // check type vi)
                        // dcb                        
                        // abc
                        // bad
                        else if ( rows[ 0 ][ 0 ] == rows[ 2 ][ 2 ]  &&  rows[ 0 ][ 1 ] == rows[ 1 ][ 2 ]  &&  rows[ 0 ][ 2 ] == rows[ 1 ][ 1 ] )
                           return true;
                        }
                     }
               }
            }
      }
   return false;
   }
