#ifndef USET_H
#define USET_H

#include "binvec.h"
#include "unique.h"

// stores an unavoidable set of size up to 14
__declspec( align( 16 ) ) struct UnavoidableSet
   {
   // empty in the beginning
   UnavoidableSet() : _size( 0 ) { assert( sizeof( UnavoidableSet ) == 256 ); }
   
   // add all clues from s, t, and u (these must be disjoint)
   UnavoidableSet( UnavoidableSet const& s, UnavoidableSet const& t, UnavoidableSet const& u )
      {
      // make sure the three unavoidable sets s, t, u are pairwise disjoint
      assert( s.IsDisjoint( t ) );
      assert( s.IsDisjoint( u ) );
      assert( t.IsDisjoint( u ) );
      BinVec128 tmp;
      tmp.Or( s._deadcluevec[ 0 ], t._deadcluevec[ 0 ] );
      _deadcluevec[ 0 ].Or( tmp, u._deadcluevec[ 0 ] );
      _size = s._size + t._size + u._size;
      }
      
   UnavoidableSet( UnavoidableSet const& s, UnavoidableSet const& t, UnavoidableSet const& u, UnavoidableSet const& v  )
      {
      assert( s.IsDisjoint( t ) );
      assert( s.IsDisjoint( u ) );
      assert( s.IsDisjoint( v ) );
      assert( t.IsDisjoint( u ) );
      assert( t.IsDisjoint( v ) );
      assert( u.IsDisjoint( v ) );
      BinVec128 tmp1, tmp2;
      tmp1.Or( s._deadcluevec[ 0 ], t._deadcluevec[ 0 ] );
      tmp2.Or( u._deadcluevec[ 0 ], v._deadcluevec[ 0 ] );
      _deadcluevec[ 0 ].Or( tmp1, tmp2 );
      _size = s._size + t._size + u._size + v._size;
      }
      
   UnavoidableSet( UnavoidableSet const& s, UnavoidableSet const& t, UnavoidableSet const& u, UnavoidableSet const& v, UnavoidableSet const& w  )
      {
      assert( s.IsDisjoint( t ) );
      assert( s.IsDisjoint( u ) );
      assert( s.IsDisjoint( v ) );
      assert( s.IsDisjoint( w ) );
      assert( t.IsDisjoint( u ) );
      assert( t.IsDisjoint( v ) );
      assert( t.IsDisjoint( w ) );
      assert( u.IsDisjoint( v ) );
      assert( u.IsDisjoint( w ) );
      assert( v.IsDisjoint( w ) );
      BinVec128 tmp1, tmp2, tmp3;
      tmp1.Or( s._deadcluevec[ 0 ], t._deadcluevec[ 0 ] );
      tmp3.Or( u._deadcluevec[ 0 ], v._deadcluevec[ 0 ] );
      tmp2.Or( tmp3, w._deadcluevec[ 0 ] );
      _deadcluevec[ 0 ].Or( tmp1, tmp2 );
      _size = s._size + t._size + u._size + v._size + w._size;
      }
      
   // returns the no. of elements in this unavoidable set
   int GetSize() const { return _size; }

   // returns the i-th clue in this unavoidable set
   int GetClue( int i ) const
      {
      assert( i >= 0  &&  i < _size );
      return _data[ i ];
      }

   bool HasClueFast( int clue ) const
      {
      assert( clue >= 0  &&  clue < 81 );
      return _deadcluevec[ 0 ].IsBitSet( clue );
      }

   void AddClue( int clue )
      {
      assert( clue >= 0  &&  clue < 81 );
      assert( ! _deadcluevec[ 0 ].IsBitSet( clue ) );
      assert( _size < 12 );
      _deadcluevec[ 0 ].SetBit( clue );
      _size++;
      }

   // determines if this unavoidable set is disjoint from another given unavoidable set
   bool IsDisjoint( UnavoidableSet const& set ) const
      {
      BinVec128 intersect;
      intersect.And( _deadcluevec[ 0 ], set._deadcluevec[ 0 ] );
      return intersect.AreAllBitsZero();
      }

   // tests if 'set' is a subset of this unavoidable set
   bool IsSubset( UnavoidableSet const& set ) const
      {
      BinVec128 intersect;    
      intersect.And( _deadcluevec[ 0 ], set._deadcluevec[ 0 ] );
      return intersect == set._deadcluevec[ 0 ];
      }
            
   void Finalize()
      {
      memset( _data, 0, sizeof( _data ) );
      int count = 0;
      for ( int i = 0;  i < 81;  i++ )
         if ( _deadcluevec[ 0 ].IsBitSet( i ) )
            _data[ count++ ] = i;
      assert( count == _size );
      _size = count;
      _deadcluevec[ 0 ].Reset();
      _deadcluevec[ 0 ].SetBit( _data[ 0 ] );
      for ( int i = 1;  i < GetSize();  i++ )
         {
         _deadcluevec[ i ] = _deadcluevec[ i - 1 ];
         _deadcluevec[ i ].SetBit( _data[ i ] );
         }
      }

   // comparison operators needed for sorting a collection of unavoidable sets
   bool operator<( UnavoidableSet const& rhs ) const
      {
      if ( GetSize() < rhs.GetSize() )
         return true;
      else if ( GetSize() > rhs.GetSize() )
         return false;
      // both sets are of the same size, need to look at individual clues
      for ( int i = 0;  i < 81;  i++ )
         if ( HasClueFast( i )  &&  ! rhs.HasClueFast( i ) )
            return true;
         else if ( ! HasClueFast( i )  &&  rhs.HasClueFast( i ) )
            return false;
      return false;
      }
      
   bool operator==( UnavoidableSet const& rhs ) const
      {
      if ( GetSize() != rhs.GetSize() )
         return false;
      // both sets have the same no. of clues, so it suffices to
      // check if one is contained in the other
      return IsSubset( rhs );
      }
   
   bool IsUnavoidable( char const* grid ) const
      {
      char puzzle[ 82 ];
      memcpy( puzzle, grid, 81 );
      puzzle[ 81 ] = '\0';
      for ( int i = 0;  i < 81;  i++ )
         if ( HasClueFast( i ) )
            puzzle[ i ] = '0';
      return GetNumberOfCompletionsOfPuzzle( puzzle ) > 1;
      }
      
   unsigned char  _data[ 15 ];
   unsigned char  _size;
   BinVec128      _deadcluevec[ 15 ];
   };

// stores an unavoidable set of size up to 12
__declspec( align( 16 ) ) struct NewUnavoidableSet
   {
   int GetSize() const { return reinterpret_cast< short const* >( _deadcluevec )[ 6 ]; }

   // returns the i-th clue in this unavoidable set
   int GetClue( int i ) const
      {
      assert( i >= 0  &&  i < GetSize() );
      return reinterpret_cast< short const* >( _deadcluevec )[ i * 8 + 7 ] >> 7;
      }

   // returns the no. of elements in this unavoidable set that are not yet dead
   int GetRealSize( BinVec128 const& deadcluevec ) const
      {
      return _deadcluevec[ 11 ].GetWeightAndNot( deadcluevec );
      }

   void operator=( UnavoidableSet const& rhs )
      {
      assert( rhs.GetSize() <= 12 );
      short* const data = reinterpret_cast< short* >( _deadcluevec );
      memcpy( data, rhs._deadcluevec, rhs.GetSize() << 4 );
      if ( rhs.GetSize() != 12 )
         memcpy( &_deadcluevec[ 11 ], &_deadcluevec[ rhs.GetSize() - 1 ], 16 );
      data[ 6 ] = rhs.GetSize();
      data[ 14 ] = rhs.GetSize() << 4;
      data[ 22 ] = rhs.GetSize() << 8;
      data[ 30 ] = rhs.GetSize() << 12;
      for ( int i = 0;  i < rhs.GetSize();  i++ )
         data[ i * 8 + 7 ] = rhs.GetClue( i ) << 7;
      }

   BinVec128   _deadcluevec[ 12 ];
   };

#endif
